/******************************************************************************
File:      DBUF.C
Project:   RocketPort DOS API
Purpose:   Double buffering sample program
Comments:  This program shows how to use the periodic event function to double
           buffer Tx data, and how to use the receive event function to double
           buffer Rx data.  The double buffering is done in a pair of queues
           for each device, these are defined in the Q_T structure.
Operation: Install a loopback plug on port 0.  At the DOS command line type
           "DBUF."  Each time transmit data is enqueued to the device 0 the Tx
           buffer the count will be displayed. Each time data is dequeued
           from the device 0 the Rx buffer the count and Rx string will be
           displayed.
*******************************************************************************/

#include <stdio.h>
#include <process.h>
#include <stdlib.h>
#include <mem.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include "api.h"

#define TRUE  1
#define FALSE 0
#define NUMDEV 8                 /* number devices this app supports */
#define TXBUF_SIZE 1024          /* transmit buffer size */
#define RXBUF_SIZE 1024          /* receive buffer size */

void EvRxData(int);              /* function prototypes */
void EvPeriodic(void);
int EnqTxData(int,unsigned char *,int);
int DeqRxData(int,unsigned char *,int);

typedef struct                   /* transmit and receive queues */
{
   int OpenTx;                   /* TRUE if device open for Tx */
   int TxIn;                     /* index to add Tx data at */
   int TxOut;                    /* index to remove Tx data at */
   unsigned char TxBuf[TXBUF_SIZE]; /* buffer for Tx data */
   int RxIn;                     /* index to add Rx data at */
   int RxOut;                    /* index to remove Rx data at */
   unsigned char RxBuf[RXBUF_SIZE]; /* buffer to Rx data */
} Q_T;

Q_T q[NUMDEV];                  /* transmit and receive queues for each device */

/******************************************************************************
Function: main
Purpose:  Initialization, test Tx and Rx double buffering.
*/
main()
{
   int Dev;
   int Err;
   unsigned char Buf[100];
   int Cnt;

   /* Initialize controller */
   aaInstallCtrlCHandler();

   if (getenv("ROCKETCFG") == NULL)
   {
     aaAutoConfig(1, 0);  // setup 1 board at io-addr 180H, IRQ10
   }
     //aaSetConfig(10, 0x180, 0,0,0);  // setup 1 board at io-addr 180H, IRQ10
   // ...else aaInit will read in config file.

   if((Err = aaInit()) != NO_ERR)
   {
      printf("Initialization Failure %x\n",Err);
      aaExit();
      exit(1);
   }

   /* Clear queues */
   for(Dev = 0;Dev < NUMDEV;Dev++)
   {
      q[Dev].OpenTx = FALSE;
      q[Dev].TxIn = 0;
      q[Dev].TxOut = 0;
      q[Dev].RxIn = 0;
      q[Dev].RxOut = 0;
   }

   /* Set up application event functions */
   aaInstallRxEvent(EvRxData);
  /**************************************************/
   //aaInstallPeriodicEvent(EvPeriodic);
   //aaEnPeriodicEvent(TRUE);

   /* Test background transmit and receive on device 0.  A loopback
      plug can be installed on device 0 so that all transmitted data will be
      received on the same device. */
   printf("To stop test press any key\n");
   if((Err = aaOpen(0,
                    COM_TX | COM_RX,
                    COM_BAUD_38400,
                    COM_PAR_NONE,
                    COM_DATABIT_8,
                    COM_STOPBIT_1,
                    COM_FLOW_NONE,
                    COM_DEN_RDA,
                    COM_MDM_RTS | COM_MDM_DTR)) != 0)
   {
      printf("Open Failure - Device %d, Error %d\n","0",Err);
      aaExit();
      exit(1);
   }
   q[0].OpenTx = TRUE;

  /**************************************************/
   aaInstallPeriodicEvent(EvPeriodic);
   aaEnPeriodicEvent(TRUE);


   while(!kbhit())                  /* test loop */
   {
      Cnt = EnqTxData(0,
               (unsigned char *)"This string is being written to device 0",
               40);
      if(Cnt > 0)
         printf("Tx %d bytes\n",Cnt);

      delay(100);                   /* wait for loopback data to get here */
      Cnt = DeqRxData(0,Buf,RXBUF_SIZE-1); /* dequeue all Rx data available */
      Buf[Cnt] = NULL;              /* null terminate the received string */
      if(Cnt > 0)
         printf("Rx %d bytes, String = %s\n",Cnt,Buf);
   }
   getch();

   /* Exit application */
   q[0].OpenTx = FALSE;
   aaClose(0,COM_MDM_RTS | COM_MDM_DTR);
   aaExit();
   return(0);
}

/******************************************************************************
Function: EnqTxData
Purpose:  Add data to a Tx queue.
Call:     EnqTxData(Dev,Buf,Cnt)
          int Dev; Device number
          unsigned char *Buf; Buffer with data to add
          int Cnt; Count of bytes to add
Return:   int: Number of bytes added to Tx queue
*/
int EnqTxData(int Dev,unsigned char *Buf,int Cnt)
{
   int i;                           /* balance of bytes to copy after q wrap */
   int NumOpen;                     /* num bytes open in Tx buffer */
   int In;                          /* In index into Tx buffer */

   asm cli;                         /* no interrupts till done, don't want
                                       periodic event function modifying q
                                       while we are working on it */
   In = q[Dev].TxIn;                /* local copy of In index */

   /* Get number bytes open in Tx buffer */
   if((NumOpen = q[Dev].TxOut - In - 1) < 0)
      NumOpen += TXBUF_SIZE;           /* adjust for q wrap */
   if(NumOpen > Cnt)
      NumOpen = Cnt;                   /* don't move more than are incoming */
   if(NumOpen == 0)
      return(0);                       /* no room in Tx buffer */
   i = NumOpen - (TXBUF_SIZE - In);    /* i = whats left after wrap around */
   if (i < 0)
      i = 0;

   /* Copy to end of Tx buffer */
   memcpy(&q[Dev].TxBuf[In],Buf,NumOpen - i);

   /* Update In index, point to beginning of buffer if already at end of it */
   In = (In + (NumOpen - i)) % TXBUF_SIZE;

   /* Copy the rest of the buffer, if any left */
   if (i != 0)
   {
      memcpy(q[Dev].TxBuf,&Buf[NumOpen - i],i);
      In = i;
   }

   /* Update Tx queue In index */
   q[Dev].TxIn = In;
   asm sti;                            /* enable interrupts */
   return(NumOpen);
}

/******************************************************************************
Function: DeqRxData
Purpose:  Remove data from a Rx queue.
Call:     DeqRxData(Dev,Buf,Cnt)
          int Dev; Device number
          unsigned char *Buf; Buffer to take data being removed from Rx queue.
          int Cnt; Count of bytes to remove
Return:   int: Number of bytes removed from Rx queue
*/
int DeqRxData(int Dev,unsigned char *Buf,int Cnt)
{
   int i;                           /* balance of bytes to copy after q wrap */
   int Out;                         /* Out index into Rx buffer */
   int BCnt;                        /* count of bytes copied */

   asm cli;                         /* no interrupts till done, don't want
                                       periodic event function modifying q
                                       while we are working on it */
   Out = q[Dev].RxOut;              /* local copy of Out index */

   /* Get number of bytes in Rx buffer */
   if((BCnt = q[Dev].RxIn - Out) < 0)
      BCnt += RXBUF_SIZE;              /* adjust for queue wrap */
   else if(BCnt == 0)
      return(BCnt);                    /* nothing in Rx buffer */
   if(Cnt < BCnt)
      BCnt = Cnt;                      /* don't move more than asked for */
   i = BCnt - (RXBUF_SIZE - Out);      /* i = whats left after wrap around */
   if(i < 0)
      i = 0;

   /* Copy to end of Rx buffer */
   memcpy(Buf,&q[Dev].RxBuf[Out],BCnt - i);

   /* Updata Out index, point to beginning of buffer if already at end of it */
   Out = (Out + (BCnt - i)) % RXBUF_SIZE;

   /* Copy the rest of the buffer, if any left */
   if (i != 0)
   {
      memcpy(&Buf[BCnt - i],q[Dev].RxBuf,i);
      Out = i;
   }

   /* Update Rx queue Out index */
   q[Dev].RxOut = Out;
   asm sti;                            /* enable interrupts */
   return(BCnt);
}

/******************************************************************************
Function: EvRxData
Purpose:  Receive event function, read data from a serial device and add it
          to a Rx queue.
Call:     EvRxData(Dev)
          int Dev; Device number
Return:   void
*/
void EvRxData(int Dev)           /* receive event function */
{
   int i;                           /* balance of bytes to copy after q wrap */
   int NumOpen;                     /* num bytes open in Rx buffer */
   int In;                          /* In index into Rx buffer */
   int Cnt;                         /* total count of bytes read */

   In = q[Dev].RxIn;                /* local copy of In index */

   /* Get number bytes open in Rx buffer */
   if((NumOpen = q[Dev].RxOut - In - 1) < 0)
      NumOpen += RXBUF_SIZE;           /* adjust for q wrap */
   if(NumOpen == 0)
      return;                          /* no room in Rx buffer */
   i = NumOpen - (RXBUF_SIZE - In);    /* i = whats left after wrap around */
   if (i < 0)
      i = 0;

   /* Read data in up to end of Rx buffer */
   Cnt = aaRead(Dev,NumOpen - i,&q[Dev].RxBuf[In]);

   /* Update In index, point to beginning of buffer if already at end of it */
   In = (In + Cnt) % RXBUF_SIZE;

   /* Read more data if any room left at front of buffer and if device wasn't
      already emptied */
   if((i != 0) &&
      (Cnt == NumOpen - i))
   {
      In = aaRead(Dev,i,q[Dev].RxBuf);  /* read balance of data */
   }

   /* Update Rx queue In index */
   q[Dev].RxIn = In;
}

/******************************************************************************
Function: EvPeriodic
Purpose:  Periodic event function, remove data from Tx queues and write it
          to serial devices.
Call:     EvPeriodic(void)
Return:   void
*/
void EvPeriodic(void)
{
   int Dev;                         /* device number */
   int i;                           /* balance of bytes to copy after q wrap */
   int Out;                         /* Out index into Tx buffer */
   int Cnt;                         /* number of bytes to write */
   int WCnt;                        /* number of bytes actually written */

   for(Dev = 0;Dev < NUMDEV;Dev++) /* check all devs for data to Tx */
   {
      if(!q[Dev].OpenTx)               /* device not open for Tx */
         continue;
      Out = q[Dev].TxOut;              /* local copy of Out index */

      /* Get number of bytes in Tx buffer */
      if((Cnt = q[Dev].TxIn - Out) < 0)
         Cnt += TXBUF_SIZE;              /* adjust for queue wrap */
      else if(Cnt == 0)
         return;                         /* nothing in Tx buffer */
      i = Cnt - (TXBUF_SIZE - Out);      /* i = whats left after wrap around */
      if(i < 0)
         i = 0;

      /* Write data to end of Tx buffer */
      WCnt = aaWrite(Dev,Cnt - i,&q[Dev].TxBuf[Out]);

      /* Updata Out index, point to start of buffer if already at end of it */
      Out = (Out + WCnt) % TXBUF_SIZE;

      /* Write more data if any left at front of buffer and if device wasn't
         already filled */
      if((i != 0) &&
         (WCnt == Cnt - i))
      {
         Out = aaWrite(Dev,i,q[Dev].TxBuf);  /* write balance of data */
      }

      /* Update Tx queue Out index */
      q[Dev].TxOut = Out;
   }
}

